# This file has been generated


build_system = 'GitHub Actions'
build_id = '1780300523'
build_number = '259'
build_version = None
job_id = '1780300523_259'
job_url = 'https://github.com/epcpower/pm/actions/runs/1780300523'
