# This file was automatically generated by sub\epyqlib\epyqlib\cli\generateversion.py
__version_build_time__ = '2022-02-01T19:34:15.351056'

__version__ = '0.0.0'
__sha__ = '968bf51d91800817193a04d0bb5c5354e7fcc1d6'
__revision__ = '968bf51d91800817193a04d0bb5c5354e7fcc1d6'
